/**
 *
 * Keystone20 [09/26/2012]
 *
 * processing 2.x compatible version of the Keystone library by David Bouchard
 *
 * (c) 2012 Rolf van Gelder - CAGE Web Design
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package rvg.keystone20;

import java.awt.Point;
import java.awt.geom.Point2D;

import javax.media.jai.PerspectiveTransform;
import javax.media.jai.WarpPerspective;

import processing.core.PApplet;
import processing.core.PGraphics;
import processing.core.PImage;
import processing.core.PVector;


/**
 * A simple Corner Pin keystoned surface. The surface is a quad mesh that can be
 * skewed to an arbitrary shape by moving its four corners.
 * 
 * September-2011 
 * 
 * Added JAI library for keystone calculus (@edumo)
 * 
 */
public class CornerPinSurface implements Draggable {

	PApplet parent;

	MeshPoint[] mesh;

	float x;
	float y;
	float clickX;
	float clickY;

	int res;

	final int TL; // top left
	final int TR; // top right
	final int BL; // bottom left
	final int BR; // bottom right

	int w;
	int h;

	int gridColor;
	int controlPointColor;
	
	//Jai class for keystone calculus

	WarpPerspective warpPerspective = null;

	/**
	 * @param parent
	 *            The parent applet -- used for default rendering mode
	 * @param w
	 *            The surface's width, in pixels
	 * @param h
	 *            The surface's height, in pixels
	 * @param res
	 *            The surface's grid resolution
	 */
	CornerPinSurface(PApplet parent, int w, int h, int res) {

		this.parent = parent;

		this.w = w;
		this.h = h;

		res++;
		this.res = res;

		// initialize the point array
		mesh = new MeshPoint[res * res];
		for (int i = 0; i < mesh.length; i++) {
			float x = (i % res) / (float) (res - 1);
			float y = (i / res) / (float) (res - 1);
			mesh[i] = new MeshPoint(this, x * w, y * h, x * w, y * h);
		}

		// indices of the corner points
		TL = 0 + 0; // x + y
		TR = res - 1 + 0;
		BL = 0 + (res - 1) * (res);
		BR = res - 1 + (res - 1) * (res);

		// make the corners control points
		mesh[TL].setControlPoint(true);
		mesh[TR].setControlPoint(true);
		mesh[BL].setControlPoint(true);
		mesh[BR].setControlPoint(true);

		calculateMesh();

		this.gridColor = 128;
		this.controlPointColor = 0xFF00FF00;
	}

	/**
	 * @return The surface's mesh resolution, in number of "tiles"
	 */
	public int getRes() {
		// The actual resolution is the number of tiles, not the number of mesh
		// points
		return res - 1;
	}

	/**
	 * Renders and applies keystoning to the image using the parent applet's
	 * renderer.
	 */
	public void render(PImage texture) {
		render(parent.g, texture);
	}

	/**
	 * Renders and applies keystoning to the image using a specific renderer.
	 */
	public void render(PGraphics g, PImage texture) {
		g.pushMatrix();
		g.translate(x, y);
		if (Keystone20.calibrate)
			g.stroke(gridColor);
		else
			g.noStroke();
		g.fill(255);
		g.beginShape(PApplet.QUADS);
		g.texture(texture);
		for (int x = 0; x < res - 1; x++) {
			for (int y = 0; y < res - 1; y++) {
				MeshPoint mp;
				mp = mesh[(x) + (y) * res];
				g.vertex(mp.x, mp.y, mp.u, mp.v);
				mp = mesh[(x + 1) + (y) * res];
				g.vertex(mp.x, mp.y, mp.u, mp.v);
				mp = mesh[(x + 1) + (y + 1) * res];
				g.vertex(mp.x, mp.y, mp.u, mp.v);
				mp = mesh[(x) + (y + 1) * res];
				g.vertex(mp.x, mp.y, mp.u, mp.v);
			}
		}
		g.endShape(PApplet.CLOSE);

		if (Keystone20.calibrate)
			renderControlPoints(g);

		g.popMatrix();
	}

	/**
	 * This function will give you the position of the mouse in the surface's
	 * coordinate system.
	 * 
	 * @return The transformed mouse position
	 */

	public PVector getTransformedCursor(int cx, int cy) {
		Point2D point = warpPerspective.mapSourcePoint(new Point(cx - (int) x,
				cy - (int) y));
		return new PVector((int) point.getX(), (int) point.getY());
	}
	

	public PVector getTransformedMouse() {
		return getTransformedCursor(parent.mouseX, parent.mouseY);
	}

	
	// 2d cross product
	private float cross2(float x0, float y0, float x1, float y1) {
		return x0 * y1 - y0 * x1;
	}

	/**
	 * Draws targets around the control points
	 */
	private void renderControlPoints(PGraphics g) {
		g.stroke(controlPointColor);
		g.noFill();
		for (int i = 0; i < mesh.length; i++) {
			if (mesh[i].isControlPoint()) {
				g.ellipse(mesh[i].x, mesh[i].y, 30, 30);
				g.ellipse(mesh[i].x, mesh[i].y, 10, 10);
			}
		}
	}

	/**
	 * Sets the grid used for calibration's color
	 */
	public void setGridColor(int newColor) {
		gridColor = newColor;
	}

	/**
	 * Sets the control points color
	 */
	public void setControlPointsColor(int newColor) {
		controlPointColor = newColor;
	}

	/**
	 * @param x
	 *            x-coordinate
	 * @param y
	 *            y-coordinate
	 */
	Draggable select(float x, float y) {
		// first, see if one of the control points are selected
		x -= this.x;
		y -= this.y;
		for (int i = 0; i < mesh.length; i++) {
			if (PApplet.dist(mesh[i].x, mesh[i].y, x, y) < 30
					&& mesh[i].isControlPoint())
				return mesh[i];
		}

		// then, see if the surface itself is selected
		if (isMouseOver()) {
			clickX = x;
			clickY = y;
			return this;
		}
		return null;
	}

	/**
	 * Returns true if the mouse is over this surface, false otherwise.
	 */
	public boolean isMouseOver() {
		if (isPointInTriangle(parent.mouseX - x, parent.mouseY - y, mesh[TL],
				mesh[TR], mesh[BL])
				|| isPointInTriangle(parent.mouseX - x, parent.mouseY - y,
						mesh[BL], mesh[TR], mesh[BR]))
			return true;
		return false;
	}

	/**
	 * Used for mouse selection of surfaces
	 */
	private boolean isPointInTriangle(float x, float y, MeshPoint a,
			MeshPoint b, MeshPoint c) {
		// http://www.blackpawn.com/texts/pointinpoly/default.html
		PVector v0 = new PVector(c.x - a.x, c.y - a.y);
		PVector v1 = new PVector(b.x - a.x, b.y - a.y);
		PVector v2 = new PVector(x - a.x, y - a.y);

		float dot00 = v0.dot(v0);
		float dot01 = v1.dot(v0);
		float dot02 = v2.dot(v0);
		float dot11 = v1.dot(v1);
		float dot12 = v2.dot(v1);

		// Compute barycentric coordinates
		float invDenom = 1 / (dot00 * dot11 - dot01 * dot01);
		float u = (dot11 * dot02 - dot01 * dot12) * invDenom;
		float v = (dot00 * dot12 - dot01 * dot02) * invDenom;

		// Check if point is in triangle
		return (u > 0) && (v > 0) && (u + v < 1);
	}

	/**
	 * Interpolates the position of the points in the mesh according to the 4
	 * corners TODO: allow for abritrary control points, not just the four
	 * corners
	 */

	protected void calculateMesh() {

		PerspectiveTransform transform = PerspectiveTransform.getQuadToQuad(0,
				0, w, 0, w, h, 0,
				h, // source to
				mesh[TL].x, mesh[TL].y, mesh[TR].x, mesh[TR].y, mesh[BR].x,
				mesh[BR].y, mesh[BL].x, mesh[BL].y); // dest

		warpPerspective = new WarpPerspective(transform);
		
		float xStep = (float) w / (res - 1);
		float yStep = (float) h / (res - 1);

		for (int i = 0; i < mesh.length; i++) {

			if (TL == i || BR == i || TR == i || BL == i)
				continue;

			float x = i % res;
			float y = i / res;
			
			x *= xStep;
			y *= yStep;

			Point2D point = warpPerspective.mapDestPoint(new Point((int) x,
					(int) y));
			mesh[i].x = (float) point.getX();
			mesh[i].y = (float) point.getY();
		}
	}

	private void calculateMeshOld() {

		for (int i = 0; i < mesh.length; i++) {
			int x = i % res;
			int y = i / res;
			float fX = (float) x / (res - 1);
			float fY = (float) y / (res - 1);
			MeshPoint bot = mesh[TL].interpolateTo(mesh[TR], fX);
			MeshPoint top = mesh[BL].interpolateTo(mesh[BR], fX);
			mesh[i].interpolateBetween(bot, top, fY);
		}
	}

	/**
	 * @param x
	 *            x-coordinate
	 * @param y
	 *            y-coordinate	 
	 * 
	 *            This moves the surface according to the offset from where the
	 *            mouse was pressed when selecting the surface.
	 */
	public void moveTo(float x, float y) {
		this.x = x - clickX;
		this.y = y - clickY;
	}

	/**
	 * @param lines
	 *            array with the lines of the text file
	 * 
	 *            Populates values from a text file
	 */
	void loadLayout(String[] lines)
	{
		String[] values = null;
		
		// s;20;0.0;0.0;400;300 - type;res;x;y;w;h
		values = parent.split(lines[0],';');
		this.x = Float.valueOf(values[2]).floatValue();
		this.y = Float.valueOf(values[3]).floatValue();
		for(int i=1; i<lines.length; i++)
		{
			// p;440;762.0;551.0;300.0;400.0 - type;x;y;u;v
			values = parent.split(lines[i],';');
			MeshPoint mp = mesh[Integer.parseInt(values[1])];
			mp.x = Float.valueOf(values[2]).floatValue();
			mp.y = Float.valueOf(values[3]).floatValue();
			mp.u = Float.valueOf(values[4]).floatValue();
			mp.v = Float.valueOf(values[5]).floatValue();
			mp.setControlPoint(true);
		}
		calculateMesh();
	}
}
