/**
 *
 * Keystone20 [09/26/2012]
 *
 * processing 2.x compatible version of the Keystone library by David Bouchard
 *
 * (c) 2012 Rolf van Gelder - CAGE Web Design
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

package rvg.keystone20;

import java.awt.event.MouseEvent;
import java.io.IOException;
import java.util.ArrayList;

import processing.core.*;
import processing.core.PApplet;
import processing.core.PGraphics;

/**
 * This class manages the creation and calibration of keystoned surfaces.
 * 
 * To move and warp surfaces, place the Keystone20 object in calibrate mode. It catches mouse events and 
 * allows you to drag surfaces and control points with the mouse.
 *
 * The Keystone20 object also provides load/save functionality, once you've calibrated the layout to 
 * your liking. 
 * 
 * Version: 003
 */
public class Keystone20 {

	public final String VERSION = "003";

	PApplet parent;

	// Container for the surfaces
	ArrayList<CornerPinSurface> surfaces;

	Draggable dragged;

	// Calibration mode is application-wide, so I made this flag static
	// there should only be one Keystone object around anyway
	static boolean calibrate;

	/**
	 * @param parent
	 *            applet
	 */
	public Keystone20(PApplet parent) {
	
		this.parent = parent;
		this.parent.registerMouseEvent(this);

		surfaces = new ArrayList<CornerPinSurface>();
		dragged = null;
	}

	/**
	 * Creates and registers a new corner pin keystone surface. 
	 * 
	 * @param w width
	 * @param h height
	 * @param res resolution (number of tiles per axis)
	 * @return CornerPinSurface
	 */
	public CornerPinSurface createCornerPinSurface(int w, int h, int res) {
		CornerPinSurface s = new CornerPinSurface(parent, w, h, res);
		surfaces.add(s);
		return s;
	}

	/**
	 * Starts the calibration mode. Mouse events will be intercepted to drag surfaces 
	 * and move control points around.
	 */
	public void startCalibration() {
		calibrate = true;
	}
	
	/**
	 * Stops the calibration mode
	 */
	public void stopCalibration() {
		calibrate = false;
	}
	
	/**
	 * Toggles the calibration mode
	 */
	public void toggleCalibration() {
		calibrate = !calibrate;
	}
	
	public boolean isCalibrating() {
		return calibrate;
	}

	/**
	 * Returns the version of the library.
	 * 
	 * @return String
	 */
	public String version() {
		return VERSION;
	}
	
	/**
	 * Saves the layout to a text file.
	 *
	 * @param filename of the text file to save to
	 */
	public void save(String filename) {
		int lineCounter = 1;
		CornerPinSurface surface = surfaces.get(0);
		
		// Determine the size of the array = surface + number of control points
		int controlPoints = 1;
		for (int i=0; i < surface.mesh.length; i++)
			if (surface.mesh[i].isControlPoint()) controlPoints++;
			
		String[] lines = new String[controlPoints];
		
 		for (CornerPinSurface s : surfaces) {
			lines[0] = "s;" + Integer.toString(s.getRes()) + ";"+ Float.toString(s.x) + ";" + Float.toString(s.y) + ";" + Integer.toString(s.w) + ";" + Integer.toString(s.h);
 			for (int i=0; i < s.mesh.length; i++) {
				if (s.mesh[i].isControlPoint()) {
					lines[lineCounter] = "p;" + Integer.toString(i) + ";" + Float.toString(s.mesh[i].x) + ";" + Float.toString(s.mesh[i].y) + ";" + Float.toString(s.mesh[i].u) + ";" + Float.toString(s.mesh[i].v);
					lineCounter++;
				}
			}
		}
		
		parent.saveStrings("data/"+filename, lines);

		PApplet.println("Keystone20: layout saved to " + filename);
	}
	
	/**
	 * Saves the current layout to "keystone20.txt"
	 */
	public void save() {
		save("keystone20.txt");
	}

	/**
	 * Loads a saved layout from a given text file
	 *
	 * @param filename of the text file to read from	 
	 */
	public void load(String filename) {
		String[] surfaceArray = null;
		
		// Load the layout text file
		try {
			surfaceArray = parent.loadStrings(filename);
		} catch (Exception e) {
			PApplet.println(e.getStackTrace());
		}
		
		// Stuff values from the file into the surface object
		surfaces.get(0).loadLayout(surfaceArray);

		PApplet.println("Keystone20: layout loaded from " + filename);
	}
	
	/**
	 * Loads a saved layout from "keystone20.txt"
	 */
	public void load() {
		load("keystone20.txt");
	}

	/**
	 * Mouse event handler
	 */
	public void mouseEvent(MouseEvent e) {
		
		// ignore input events if the calibrate flag is not set
		if (!calibrate)
			return;

		int x = e.getX();
		int y = e.getY();

		switch (e.getID()) {
			case MouseEvent.MOUSE_PRESSED:
				CornerPinSurface top = null;
				// navigate the list backwards, as to select 
				for (int i=surfaces.size()-1; i >= 0; i--) {
					// CornerPinSurface s = (CornerPinSurface)surfaces.get(i);
					CornerPinSurface s = surfaces.get(i);
					dragged = s.select(x, y);
					if (dragged != null) {
						top = s;
						break;
					}
				}
				break;

			case MouseEvent.MOUSE_DRAGGED:
				if (dragged != null)
					dragged.moveTo(x, y);
				break;

			case MouseEvent.MOUSE_RELEASED:
				dragged = null;
				break;
		}
	} // mouseEvent(MouseEvent e)
} // public class Keystone20